"""
@file
@brief Shortcuts to tools.
"""

from .asv_options_helper import get_opset_number_from_onnx, get_ir_version_from_onnx
from .code_helper import change_style
from .speed_measure import measure_time
